#!/usr/bin/env python3
"""
Cheap Pie Hardware Abstraction Layer
"""
#
# -*- coding: utf-8 -*-
# this file is part of cheap_pie, a python tool for chip validation
# author: Marco Merlin
# email: marcomerli@gmail.com

import os
import hickle as hkl

import cheap_pie.tools.search  # pylint: disable=W0611
from cheap_pie.tools.hal2doc import hal2doc, int2hexstr, hexstr2int


class CpHal():
    """
    Cheap Pie Hardware Abstraction Layer
    """
    regs = []
    hif = None

    def __init__(self, regs):
        if isinstance(regs, CpHal):
            # this allows generating CpHal super-classes from a CpHal instance
            self.regs = regs.regs
        else:
            self.regs = regs
            if len(regs) > 0:
                self.hif = regs[0].hif
            else:
                print('# WARNING: no register defined!')

    def __len__(self):
        return len(self.regs)

    def __iter__(self):
        return self.regs.__iter__()

    def __next__(self):
        return self.regs.next()

    def __getitem__(self, idx):
        if isinstance(idx, int):
            return self.regs[idx]
        if isinstance(idx, str):
            return self.regs._asdict()[idx]
        assert False, 'Unsupported indexing!'

    def __setitem__(self, idx, value):
        if isinstance(idx, int):
            return self.regs[idx].setreg(value)
        if isinstance(idx, str):
            return self.regs._asdict()[idx].setreg(value)
        assert False, 'Unsupported indexing!'

    def __repr__(self):
        return 'CP_HAL: %d registers' % len(self.regs)  # pylint: disable=C0209

## search methods ###########################################################

    def search_bitfield(self, field, case_sensitive=False):
        """
        Search bitfields which name contains a specified string
        """
        return cheap_pie.tools.search.bitfield(self.regs, field, case_sensitive=case_sensitive)  # pylint: disable=E0601

    def search_register(self, reg, case_sensitive=False):
        """
        Search registers which name contains a specified string
        """
        return cheap_pie.tools.search.register(self.regs, reg, case_sensitive=case_sensitive)

    def search_address(self, address, mask='0xFFFFFFFF'):
        """
        Search register matching a specified address
        """
        return cheap_pie.tools.search.address(self.regs, address, mask=mask)

    def to_docx(self, *args):
        """
        Generate a .docx document describing an Hardware Abstraction Layer
        """
        hal2doc(self.regs, *args)

## dump methods #############################################################
    def regs2dict(self):
        """
        Converts an hal structure into a dictionary in the form:
        dictval = {
            'reg1': val1,
            'reg2': val2,
        }
        """
        outdict = {}
        for reg, val in self.regs._asdict().items():
            outdict[reg] = val.getreg()
        return outdict

    def dump(self, fname='dump.hkl', regs_dict=None):
        """
        Dump the value of all registers in a .hkl file.
        Inputs:
            - fname: .hkl or .txt output file
            - regs_dict: specify a dictionary of register values.
                registers will be read from chip, if not provided.
        """

        if regs_dict is None:
            # read all registers
            regs_dict = self.regs2dict()

        # detect input file type from extension
        _, file_extension = os.path.splitext(fname)
        assert file_extension in ['.txt', '.hkl']

        if file_extension == '.txt':
            self.dump2text(f1name=fname, field1=regs_dict)
        else:
            hkl.dump(regs_dict, fname, compression='gzip')

    def dump_load(self, fname='dump.hkl'):
        """
        Load a dumped file and returns a dictionary of registers
        """
        _, file_extension = os.path.splitext(fname)
        assert file_extension in ['.txt', '.hkl']

        if file_extension == '.txt':
            _, fields = self.text2dump(fname)
        else:
            fields = hkl.load(fname)

        return fields

    def dump_diff(self, f1name='dump.hkl', f2name='dump2.hkl', width=60):
        """
        Perform a diff on two dumped .hkl files.
        """
        fmtstr = '%%%ds |%%%ds' % (width, width)  # pylint: disable=C0209

        field1 = self.dump_load(f1name)
        field2 = self.dump_load(f2name)

        # create a header with filenames
        outstrlist = []
        for reg, val in field1.items():
            # print(f'reg:{reg}, val: {val}')
            if len(reg) > 0: # just check reg that have a name
                if not field2[reg] == val:
                    f1regstr = self[reg].__repr__(val).split(  # pylint: disable=C2801
                        '\n')
                    f2regstr = self[reg].__repr__(field2[reg]).split(  # pylint: disable=C2801
                        '\n')

                    for idx in range(len(f1regstr)):  # pylint: disable=C0200
                        if not f1regstr[idx] == f2regstr[idx]:
                            outstrlist.append(fmtstr %
                                            (f1regstr[idx], f2regstr[idx]))

        # print output
        if len(outstrlist) > 0:
            headerstr = fmtstr % (f1name, f2name)
            outstrlist.insert(0, headerstr)
            for line in outstrlist:
                print(line)
        else:
            print('No differences found!!!')

        return outstrlist

    def dump2text(self, f1name='dump.hkl', field1=None,
                  width=60, save_en=True, print_en=False, header_en=False,
                  byval=True, nbits_addr=32, nbits_val=32
                  ):  # pylint: disable=R0913,R0914
        """
        Convert a dumped .hkl file, or a dumped dictionary into a text file representation.
        Inputs:
            f1name: .hkl file containing dump dictionary, also defines .txt file output name
            field1: input dictionary of dumped values to save.
                    .hkl file specified by f1name must exist, if this is not specified
            width: number of columns reserved for register and bitfield name
            save_en: save output to text file (default: True)
            print_en: print output to shell (default: False)
            byval: dump in hex ascii format. ie ADD43550 C0D3C057 (default: True)
            nbits_addr: number of address bits for byval dump (default: 32)
            nbits_val: number of value bits for byval dump (default: 32)
        """

        if not isinstance(field1, dict):
            assert os.path.isfile(f1name)
            field1 = hkl.load(f1name)

        fmtstr = '%%%ds' % width  # pylint: disable=C0209

        # generate register description list
        outstrlist = []
        for reg, val in field1.items():
            if byval:
                f1regstr = '%s %s\n' % (  # pylint: disable=C0209
                    int2hexstr(self[reg].addr, nbits_addr/4),
                    int2hexstr(val, nbits_val / 4)
                )
                outstrlist.append(f1regstr)
            else:
                f1regstr = self[reg].__repr__(val).split(  # pylint: disable=C2801
                    '\n')
                for idx in range(len(f1regstr)):  # pylint: disable=C0200
                    outstrlist.append(fmtstr % f1regstr[idx])

        fname_wo_ext, _ = os.path.splitext(f1name)

        if header_en:
            # create a header with filenames
            headerstr = f'{f1name}\n'
            outstrlist.insert(0, headerstr)

        if save_en:
            outfname = fname_wo_ext + '.txt'
            file_handler = open(outfname, "w", encoding='utf8') # pylint: disable=R1732
            print(f'Output file: {outfname}')

        # print output
        if len(outstrlist) > 0:
            for line in outstrlist:
                if print_en:
                    print(line)
                if save_en:
                    file_handler.write(line)

        # close file
        if save_en:
            file_handler.close()

        # return outstrlist

    def text2dump(self, fname='dump.txt', save_en=True):
        """
        Convert a dumped .txt file, in a .hkl file or a dictionary.
        Inputs:
            f1name: .txt file containing dump address and value in the form:
                ADD43550 C0D3C057
        """

        assert os.path.isfile(fname)

        regs_dict = {}
        with open(fname, 'r', encoding='utf8') as file_handler:
            for line in file_handler:
                try:
                    line = line.strip()
                    if len(line) > 0:
                        addrstr, valstr = line.split()
                        # print(f'addr: {addrstr}, val: {valstr}')

                        addr = hexstr2int(addrstr)
                        val = hexstr2int(valstr)

                        regname = self.search_address(addr)
                        # print(regname)

                        regs_dict[regname] = val
                except Exception as err: # pylint: disable=W0718
                    print(f"Unable to process line: <{line}>",repr(err))

        if save_en:
            fname_wo_ext, _ = os.path.splitext(fname)
            outfname = fname_wo_ext + '.hkl'
            self.dump(fname=outfname, regs_dict=regs_dict)
        else:
            outfname = ''

        return outfname, regs_dict


class CpHalSuper(CpHal):
    """ Just a dummy class for test """

    def super_method(self):
        """ Just a dummy method of super-class CpHalSuper """
        print('Super Method!!!')


def test_cp_hal_to_docx():
    """
    Test Function for Cheap Pie HAL to .docx
    """
    from parsers.cp_parsers_wrapper import cp_parsers_wrapper  # pylint: disable=C0415,C0413,E0401
    from cheap_pie_core.cp_cli import cp_cli                   # pylint: disable=C0415,C0413,E0401

    prms = cp_cli(['-t', 'dummy', '-rf', 'my_subblock.xml', '-fmt', 'ipxact'])
    hal = cp_parsers_wrapper(prms)
    hal.to_docx()


def test_cp_hal():  # pylint: disable=R0914,R0915
    """
    Test Function for Cheap Pie Hardware Abstraction Layer
    """
    from ast import literal_eval                                        # pylint: disable=C0415
    from parsers.cp_parsers_wrapper import cp_parsers_wrapper  # pylint: disable=C0415,E0401
    from transport.cp_dummy_transport import CpDummyTransport  # pylint: disable=C0415,E0401
    from cheap_pie_core.cp_cli import cp_cli                  # pylint: disable=C0415,E0401

    print('# hal initialize')
    prms = cp_cli(['-t', 'dummy'])
    hal = cp_parsers_wrapper(prms, hif=CpDummyTransport())

    print('# hal test register methods...')
    print('# hal hex assignement')
    inval = "0xFFFFFFFF"
    hal.regs.ADC_ANA_CTRL.setreg(inval)
    retval = hex(hal.regs.ADC_ANA_CTRL.getreg())
    print(retval)
    assert literal_eval(inval) == literal_eval(retval)

    print('# hal decimal assignement')
    inval = 2
    hal.regs.ADC_ANA_CTRL.setreg(inval)
    retval = hal.regs.ADC_ANA_CTRL.getreg()
    assert inval == retval

    print('# hal reg representation')
    hal.regs.ADC_ANA_CTRL  # pylint: disable=W0104

    print('# hal reg display')
    hal.regs.ADC_ANA_CTRL.display()

    print('# hal reg print')
    print(hal.regs.ADC_ANA_CTRL)

    print('# hal reg str')
    print(str(hal.regs.ADC_ANA_CTRL))

    print('# hal test bitfield methods...')
    print('# hal bitfield display')
    hal.regs.ADC_ANA_CTRL.bitfields.ADC_BM.display()
    print('# hal bitfield display with value')
    hal.regs.ADC_ANA_CTRL.bitfields.ADC_BM.display(2)
    print('# hal bitfield setbit')
    hal.regs.ADC_ANA_CTRL.bitfields.ADC_BM.setbit(inval)
    retval = hal.regs.ADC_ANA_CTRL.bitfields.ADC_BM.getbit()
    assert inval == retval

    print('# hal subscriptable interface...')
    print('# hal[0]')
    print(hal[0])

    print('# hal[0][0]')
    print(hal[0][0])

    print("# hal['ADC_ANA_CTRL']")
    print(hal['ADC_ANA_CTRL'])

    print('# hal dict-based reg assignement')
    regval = 1
    hal['ADC_ANA_CTRL'] = regval
    assert hal.regs.ADC_ANA_CTRL.getreg() == regval

    print('# hal dict-based bitfield assignement')
    fieldval = 2
    hal['ADC_ANA_CTRL']['ADC_BM'] = fieldval
    assert hal.regs.ADC_ANA_CTRL.bitfields.ADC_BM.getbit() == fieldval

    print('# hal dict-based assignement in single register write')
    dither = 1
    chop = 1
    inv = 1
    hal['ADC_ANA_CTRL'] = {'DITHER_EN': dither,
                           'CHOP_EN': chop, 'INV_CLK': inv}
    assert hal.regs.ADC_ANA_CTRL.bitfields.DITHER_EN.getbit() == dither
    assert hal.regs.ADC_ANA_CTRL.bitfields.CHOP_EN.getbit() == chop
    assert hal.regs.ADC_ANA_CTRL.bitfields.INV_CLK.getbit() == inv

    print('# hal search_register')
    searchreg = 'ADC_ANA_CTRL'
    reglist = hal.search_register(searchreg)
    reg = reglist[0]
    print(reg)
    assert str(reg) == searchreg

    print('# hal search_bitfield')
    searchfield = 'ADC_BM'
    fieldlist = hal.search_bitfield(searchfield)
    print(fieldlist)
    field = fieldlist[0]
    print(field)
    assert searchfield in field

    print('# hal search_address')
    searchaddr = '0x4000702c'
    reg = hal.search_address(searchaddr)
    print(f'# Register {reg}')
    addr = hex(hal[reg].addr)
    print(f'# Address: {addr}')
    assert addr == searchaddr

    print('# hal search_address with mask')
    searchaddr = '0xF000702c'
    searchmask = '0x0FFFFFFF'
    reg = hal.search_address(searchaddr, mask=searchmask)
    print(f'# Register {reg}')
    addr = hex(hal[reg].addr)
    print(f'# Address: {addr}')
    assert (literal_eval(addr) & literal_eval(searchmask) ==
            literal_eval(searchaddr) & literal_eval(searchmask))
    reg = hal.search_address(searchaddr)
    assert reg == ''

    print('# hal dump')
    dump1 = 'dump1.hkl'
    dump2 = 'dump2.hkl'
    hal.dump(dump1)
    hal['ADC_ANA_CTRL']['ADC_BM'] = 3
    hal.dump(dump2)
    diff = hal.dump_diff(dump1, dump2)
    print(diff)
    assert len(diff) == 2

    print('# hal regs2dict')
    mydict = hal.regs2dict()
    assert isinstance(mydict, dict)

    print('# hal dump text')
    txt_dump = 'tdump.txt'
    hal.dump(txt_dump)
    hkl_dump, _ = hal.text2dump(txt_dump)
    assert len(hal.dump_diff(hkl_dump, dump2)) == 0

    print('# CpHalSuper inheritance')
    hal_super = CpHalSuper(hal)
    hal_super.super_method()


if __name__ == '__main__':
    test_cp_hal()



#!/usr/bin/env python3
""" Cheap Pie module for native .svd files parser """
# this file is part of cheap_pie, a python tool for chip validation
# author: Marco Merlin
# email: marcomerli@gmail.com

# parser build for CMSIS-SVD xml file format
# https://www.keil.com/pack/doc/CMSIS/SVD/html/index.html
#
# simplified manual initialization
#
# import untangle
# hal = untangle.parse('QN908XC.xml')
# from cbitfield import cp_bitfield
# from cp_register import cp_register
# ADC_ANA_CTRL = cp_register("ADC_ANA_CTRL","0x4000702C",
#                            "ADC core and reference setting regsiter" , hif )
# ADC_BM = cp_bitfield("ADC_BM","0x4000702C","ADC_ANA_CTRL",3,0, "ADC bias current selection." ,hif)

from ast import literal_eval
import untangle  # for parsing xml

from cheap_pie.cheap_pie_core.cp_builder import CpHalBuilder  # pylint: disable=C0413,E0401
from cheap_pie.cheap_pie_core.cp_cli import cp_devices_fname


def svd_parse(fname, vendor=None, hif=None, base_address_offset="0x00000000"):  # pylint: disable=W0613
    """ Cheap Pie native parser for .svd files """
    ## read input file ########################################################
    svd = untangle.parse(fname)

    ## loop over lines ########################################################
    cpb = CpHalBuilder(hif)

    for periph in svd.device.peripherals.peripheral:  # pylint: disable=R1702
        # print(periph.name.cdata)

        base_addr_str = periph.baseAddress.cdata
        base_address = literal_eval(base_addr_str)

        if hasattr(periph, 'registers'):
            for reg in periph.registers.register:
                if hasattr(reg.name, 'cdata'):
                    # new register
                    addr_str = reg.addressOffset.cdata
                    regaddr = (
                        literal_eval(addr_str) +
                        base_address +
                        literal_eval(base_address_offset)
                    )

                    cpb.reg_open(
                        regname=f'{periph.name.cdata}_{reg.name.cdata}',
                        regaddr=regaddr,
                        comments=reg.description.cdata,
                    )

                    if hasattr(reg, 'fields'):
                        for field in reg.fields.field:
                            if not field is None:
                                # Create new field class
                                cpb.newfield( # pylint: disable=R0801
                                    regfield=field.name.cdata,
                                    width=field.bitWidth.cdata,
                                    offset=field.bitOffset.cdata,
                                    comments=field.description.cdata,
                                )

    # convert output dictionary into structure
    return cpb.out()


def test_svd_parse(argv=[]):  # pylint: disable=W0102
    """ test function for .svd parser """

    if len(argv) > 1:
        fname = argv[1]
    else:
        fname = cp_devices_fname("QN908XC.svd")

    hal = svd_parse(fname)
    assert len(hal) > 0

    return hal


if __name__ == '__main__':
    import sys
    print(svd_parse(sys.argv))



#!/usr/bin/env python3

""" Convert between different .xml formats using .xslt rules
Examples of conversion between different IP-XACT versions
"""

import sys
import argparse
import os
import difflib

import lxml.etree as ET
from cheap_pie.cheap_pie_core.cp_cli import cp_devices_fname, cp_root_dir


def xml_xslt_parse(args=[]):  # pylint: disable=W0102
    """ Parsing function for xml converter """
    parser = argparse.ArgumentParser(description='Apply XSLT to .xml file')
    parser.add_argument("-in", "--input", help=".xml input",
                        action='store', type=str, default=cp_devices_fname("generic_example.xml"))
    parser.add_argument("-xslt", "--xslt", help=".xslt file",
                        action='store', type=str,
                        default=os.path.join(
                            cp_root_dir(),
                            "parsers/ipxact_rules/from1685_2009_to_1685_2014.xsl"
                        )
                        )
    parser.add_argument("-out", "--output", help=".xml output",
                        action='store', type=str, default="output.xml")
    return parser.parse_args(args)


def xml_xslt(xmlin, xslt, xmlout):
    """ Convert between different .xml formats using .xslt rules """
    dom = ET.parse(xmlin)  # pylint: disable=I1101
    xslt = ET.parse(xslt)  # pylint: disable=I1101
    transform = ET.XSLT(xslt)   # pylint: disable=I1101
    newdom = transform(dom)
    infile = ET.tostring(newdom, pretty_print=True).decode( # pylint: disable=I1101
        'utf-8')
    # print(infile)
    if os.path.isfile(xmlout):
        os.remove(xmlout)
    with open(xmlout, 'a', encoding='utf-8') as outfile:
        outfile.write(infile)


def compare(fromfile, tofile):
    """ Compare two text files """
    with open(fromfile, encoding='utf-8') as ffh:
        fromlines = ffh.readlines()
    with open(tofile, encoding='utf-8') as tfh:
        tolines = tfh.readlines()
    diff = difflib.unified_diff(
        fromlines, tolines, fromfile=fromfile, tofile=tofile)

    difflines = []
    for line in diff:
        difflines.append(line)
        # print(line)

    assert len(difflines) > 0


def test_xml_xslt(args=[]):  # pylint: disable=W0102
    """ Test convert between different .xml formats using .xslt rules """
    prms = xml_xslt_parse(args)
    xml_xslt(xmlin=prms.input, xslt=prms.xslt, xmlout=prms.output)
    # diff input output
    compare(prms.input, prms.output)


if __name__ == '__main__':
    test_xml_xslt(sys.argv[1:])



#!/usr/bin/env python3
""" A dummy mockup transport module for Cheap Pie """

# -*- coding: utf-8 -*-
# this file is part of cheap_pie, a python tool for chip validation
# author: Marco Merlin
# email: marcomerli@gmail.com

from ast import literal_eval
from abc import ABCMeta


def hex_bw(val, hex_digits_width=8):
    """ converts integer value into an hex string of fixed width """
    assert isinstance(val, int)

    retstr = "{0:#0{1}x}".format( # pylint: disable=C0209
        val, hex_digits_width+2)

    #  Explanation:
    # {   # Format identifier
    # 0:  # first parameter
    # #   # use "0x" prefix
    # 0   # fill with zeroes
    # {1} # to a length of n characters (including 0x), defined by the second parameter
    # x   # hexadecimal number, using lowercase letters for a-f
    # }   # End of format identifier

    # print(retstr)

    return retstr


def numeric_input(val):
    """ Numeric input that supports string evaluation with literal_eval """
    if isinstance(val, int):
        return val
    if isinstance(val, str):
        return int(literal_eval(val))
    assert False, f'Unsupported input type: {type(val)}'


def hifread_preproc(addr):
    """ convert input of hifread into numeric """
    return numeric_input(addr)


def hifwrite_preproc(addr, val, mask='0xFFFFFFFF'):
    """ convert input of hifwrite into numeric """
    return numeric_input(addr), numeric_input(val), numeric_input(mask)


class CpDummyTransport(metaclass=ABCMeta):
    """ A transport mockup """
    mem = {}

    def hifread(self, addr="0x40000888"):
        """ Mockup for read a register """

        # make sure string format is always the same
        addr = hifread_preproc(addr)

        addrstr = hex_bw(addr)

        # check address exists, or create it
        if addrstr not in self.mem.keys():  # pylint: disable=C0201
            self.mem[addrstr] = 0

        return self.mem[addrstr]

    def hifwrite(self, addr="0x40000888", val="0x00000352", verify=True):
        """ Mockup for write a register """

        waddr, wval, _ = hifwrite_preproc(addr, val)

        self.mem[hex_bw(waddr)] = wval

        if verify:
            assert int(self.hifread(addr))==int(wval)

        return int(wval)


def test_cp_dummy(transport=CpDummyTransport()):
    """ Test transport mockup """
    print("# Test dummy transport")
    addr = 4
    val = 5
    transport.hifwrite(addr=addr, val=val)
    assert val == transport.hifread(addr=addr)


if __name__ == '__main__':
    test_cp_dummy()



#!/usr/bin/env python3
"""
Namedtuple HAL builder for Cheap Pie
"""
#
# -*- coding: utf-8 -*-
# this file is part of cheap_pie, a python tool for chip validation
# author: Marco Merlin
# email: marcomerli@gmail.com

# import sys     # pylint: disable=C0411
# import os.path # pylint: disable=C0411
# sys.path.append( os.path.join(os.path.dirname(__file__), '..') )

from cheap_pie.cheap_pie_core.cp_register import dict2namedtuple, isnamedtupleinstance, CpRegBuilder  # pylint: disable=C0413,E0401
from cheap_pie.cheap_pie_core.cp_hal import CpHal  # pylint: disable=C0413,E0401


def name_subs(regname=None):
    """ Names Substitution function for Cheap Pie parsers """

    # print(regname)
    # regname=strrep(regname,'"','')
    regname = regname.strip()
    regname = regname.replace('"', '')
    regname = regname.replace('[', '')
    regname = regname.replace(']', '')
    regname = regname.replace('%', '')
    regname = regname.replace('/', '')
    regname = regname.replace('\\', '')
    regname = regname.replace(' ', '_')
    if regname[0].isdigit():
        regname = 'M' + regname
    return regname


class CpHalBuilder():
    """ Namedtuple HAL builder for Cheap Pie """

    hif = None
    outdict = {}
    struct_register = None

    def __init__(self, hif=None):
        self.hif = hif
        # for some reason need to reset this
        self.outdict = {}

    def reg_close(self):
        """ close a register instance declaration """
        if not self.struct_register is None:
            # check
            assert isinstance(self.struct_register, CpRegBuilder)
            self.outdict[self.struct_register.regname] = self.struct_register.dictfield2struct(
            )
            self.struct_register = None

    def reg_open(self, regname, regaddr, comments=''):
        """ start a register instance declaration """
        self.reg_close()
        self.struct_register = CpRegBuilder(
            regname=name_subs(regname),
            regaddr=regaddr,
            comments=comments,
            hif=self.hif)

    def nregs(self):
        """ Return the number of registers already created """
        retval=len(self.outdict)
        if not self.struct_register is None:
            retval += 1
        return retval

    def reg_exists(self,regname):
        """ Returns true if a register with the specified name already exists """
        if not self.struct_register is None:
            if self.struct_register.regname == regname:
                return True

        for key,_ in self.outdict.items():
            if key==regname:
                return True

        return False

    def newfield(self, regfield, width, offset, comments):
        """ add a new field to current register """
        assert isinstance(self.struct_register, CpRegBuilder)
        self.struct_register.addfield(
            regfield=name_subs(regfield),
            width=width,
            offset=offset,
            comments=comments,
        )

    def out(self):
        """ returns a namedtuple that represent the register list """
        self.reg_close()
        return CpHal(dict2namedtuple(outdict=self.outdict))


def test_cp_builder():
    """ test cp_builder """
    cpb = CpHalBuilder()

    assert cpb.nregs()==0
    assert cpb.reg_exists('reg1') is False
    cpb.reg_open('reg1', 1, 'comment1')
    cpb.newfield('reg1_field1', offset=0, width=1, comments='reg1_field1')
    cpb.newfield('reg1_field2', offset=1, width=2, comments='reg1_field2')
    assert cpb.nregs()==1
    assert cpb.reg_exists('reg1') is True

    cpb.reg_open('reg2', 2, 'comment2')
    assert cpb.nregs()==2
    cpb.newfield('reg2_field1', offset=0, width=4, comments='reg2_field1')
    cpb.newfield('reg2_field2', offset=5, width=2, comments='reg2_field2')
    cpb.newfield('reg2_field3', offset=8, width=2, comments='reg2_field3')

    regfile = cpb.out()

    print('# REG1')
    print(regfile.regs.reg1)
    print('# REG2')
    print(regfile.regs.reg2)

    assert isinstance(regfile, CpHal)
    assert len(regfile) == 2

    assert isnamedtupleinstance(regfile.regs)
    assert isnamedtupleinstance(regfile.regs.reg2.bitfields)

    assert len(regfile.regs) == 2
    assert len(regfile.regs.reg1.bitfields) == 2
    assert len(regfile.regs.reg2.bitfields) == 3


if __name__ == '__main__':
    test_cp_builder()



#!/usr/bin/env python3
""" Cheap Pie Test Module """

# -*- coding: utf-8 -*-
# this file is part of cheap_pie, a python tool for chip validation
# author: Marco Merlin
# email: marcomerli@gmail.com

import unittest

if __name__ == '__main__':
    # needed if cheap_pie not installed, and running locally
    import sys
    import os.path
    sys.path.append(os.path.join(os.path.dirname(__file__), '../..'))

# Cheap Pie Core
from cheap_pie.cheap_pie_core.cbitfield import test_cp_bitfield               # pylint: disable=C0413,E0401
from cheap_pie.cheap_pie_core.cp_register import test_cp_register             # pylint: disable=C0413,E0401
from cheap_pie.cheap_pie_core.cp_builder import test_cp_builder               # pylint: disable=C0413,E0401
from cheap_pie.cheap_pie_core.cp_hal import test_cp_hal, test_cp_hal_to_docx  # pylint: disable=C0413,E0401
from cheap_pie.cheap_pie_core.cheap_pie_main import cp_main                                  # pylint: disable=C0413,E0401

# parsers
from cheap_pie.parsers.svd_parse import test_svd_parse                        # pylint: disable=C0413,E0401
from cheap_pie.parsers.svd_parse_repo import test_svd_parse_repo              # pylint: disable=C0413,E0401
from cheap_pie.parsers.ipxact_parse import test_ipxact_parse                  # pylint: disable=C0413,E0401
from cheap_pie.parsers.ipyxact_parse import test_ipyxact_parse                # pylint: disable=C0413,E0401
from cheap_pie.parsers.rdl_parse import test_rdl_parse                        # pylint: disable=C0413,E0401
from cheap_pie.parsers.cp_parsers_wrapper import test_cp_parsers_wrapper      # pylint: disable=C0413,E0401
from cheap_pie.parsers.xml_xslt import test_xml_xslt                          # pylint: disable=C0413,E0401

# tools
from cheap_pie.tools.hal2doc import test_hal2doc                              # pylint: disable=C0413,E0401
from cheap_pie.tools.search import test_search                                # pylint: disable=C0413,E0401
from cheap_pie.tools.rdl2any import test_rdl2any                              # pylint: disable=C0413,E0401

# Transport
from cheap_pie.transport.cp_dummy_transport import test_cp_dummy              # pylint: disable=C0413,E0401
from cheap_pie.transport.cp_jlink_transport import test_cp_jlink              # pylint: disable=C0413,E0401
from cheap_pie.transport.cp_pyocd_transport import test_cp_pyocd              # pylint: disable=C0413,E0401
from cheap_pie.transport.cp_esptool_transport import test_cp_esptool          # pylint: disable=C0413,E0401
from cheap_pie.transport.cp_pyverilator_transport import test_cp_pyverilator  # pylint: disable=C0413,E0401


class CheapPieMethods(unittest.TestCase):
    """ Cheap Pie Test Class """

    def test_transport(self):
        """ Cheap Pie Test Transport Method """

        # dummy for mockup
        test_cp_dummy()
        # NB: requirements_extra.txt are needed!
        test_cp_jlink()
        test_cp_pyocd()
        test_cp_esptool()
        test_cp_pyverilator()

    def test_cheap_pie_core(self):
        """ Test cheap pie core classes """
        test_cp_bitfield()
        test_cp_register()
        test_cp_builder()
        test_cp_hal()
        test_cp_hal_to_docx()
        cp_main()

    def test_parsers(self):
        """ Test cheap pie parsers """
        test_svd_parse()
        test_svd_parse_repo()
        test_ipxact_parse()
        test_ipyxact_parse()
        test_rdl_parse()
        test_xml_xslt()
        test_cp_parsers_wrapper()

    def test_tools(self):
        """ Test cheap_pie tools """
        test_hal2doc()
        test_search()
        test_rdl2any()


if __name__ == '__main__':
    unittest.main()



""" Pypi setup for cheap_pie """
from setuptools import setup, find_packages

with open("README.md", "r", encoding='utf8') as fh:
    long_description = fh.read()

setup(
    name='cheap_pie',
    version='1.1.0',
    license='Apache 2.0',
    author="Marco Merlin",
    author_email='marcomerli@gmail.com',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    description="A python tool for silicon validation.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/bat52/cheap_pie',
    keywords='python silicon validation',
    install_requires=[
          'untangle',
          'hickle',
          # parsers
          'cmsis-svd',
          'ipyxact',
          'python-docx',
          'wavedrom',
          'wavedrom-ascii',
          'pyverilator-mm',
          'peakrdl-ipxact',
          'peakrdl-uvm',
          'peakrdl-verilog',
          'peakrdl-cheader',
          # 'lxml', # for xlst ipxact conversions
          # transport layers
          #'pylink-square',
          #'pyocd',
          #'esptool',
          #'packaging' # for verilator version check
      ],
)

project_urls={
    "Source": "https://github.com/bat52/cheap_pie",
    "Tracker": "https://github.com/bat52/cheap_pie/issues"
}



#!/usr/bin/env python3
"""
Cheap Pie module for pyverilator transport
"""
import os
import argparse
import sys
from ast import literal_eval
from shutil import copyfile
import subprocess
from packaging import version

import pyverilator

from cheap_pie.transport.cp_dummy_transport import hifread_preproc, hifwrite_preproc  # pylint: disable=E0401


def cli(args):
    """ Command Line Interface for pyverilator transport class """
    parser = argparse.ArgumentParser(description='rdl2verilog pyverilator ')
    # register format options
    parser.add_argument("-f", "--fname",
                        help="register file description .v", action='store',
                        type=str, default="./devices/verilog/basic_rf.v")
    return parser.parse_args(args)


class CpPyverilatorTransport():
    """
    Cheap Pie class for pyverilator transport
    """
    sim = None

    def __init__(self, fname, gtkwave_en=False):

        # rename to .v, if .sv
        if not os.path.isfile(fname):
            assert False, f'File {fname} does not exist!'

        base, ext = os.path.splitext(fname)
        # print(ext)

        if ext == '.sv':
            print('renaming input file to .v')
            ofname = base + '.v'
            copyfile(fname, ofname)
        else:
            ofname = fname

        print(ofname)
        self.sim = pyverilator.PyVerilator.build(ofname)

        if gtkwave_en:  # still causing trouble in VSC and/or WSL
            # start gtkwave to view the waveforms as they are made
            self.sim.start_gtkwave()

            # add all the io and internal signals to gtkwave
            # sim.send_signal_to_gtkwave(sim.io)
            # sim.send_signal_to_gtkwave(sim.internals)

            # add all the io and internal signals to gtkwave
            # self.sim.send_to_gtkwave(self.sim.io)
            # self.sim.send_to_gtkwave(self.sim.internals)

            self.sim.clock.send_to_gtkwave()
            self.sim.io.resetn.send_to_gtkwave()  # pylint: disable=E1101
            self.sim.io.addr.send_to_gtkwave()
            self.sim.io.wdata.send_to_gtkwave()
            self.sim.io.rdata.send_to_gtkwave()

        self.reset_release()

    def reset_release(self):
        """ Release Reset signal """
        # tick the automatically detected clock
        self.sim.clock.tick()

        # set rst back to 0
        # sim.io.rst = 0
        self.sim.io.resetn = 0
        self.sim.clock.tick()
        self.sim.io.resetn = 1

        # check out when en = 0
        # sim.io.en = 0
        # curr_out = sim.io.out
        # sim.io is a pyverilator.Collection, accessing signals by attribute or
        # dictionary syntax returns a SignalValue object which inherits from int.
        # sim.io.out can be used just like an int in most cases, and it has extra
        # features like being able to add it to gtkwave with
        # sim.io.out.send_to_gtkwave(). To just get the int value, you can call
        # sim.io.out.value
        # print('sim.io.out = ' + str(curr_out))

        # check out when en = 1
        # sim.io.en = 1
        # curr_out = sim.io.out
        # print('sim.io.out = ' + str(curr_out))

    def hifwrite(self, addr='0x00', val='0xB16B00B5', mask='0xFFFFFFFF', verify=True):
        """ Write register """

        addr, val, mask = hifwrite_preproc(addr, val, mask)

        # write value
        self.sim.io.addr = addr
        self.sim.io.wdata = val
        self.sim.io.wmask = mask
        self.sim.io.read = 0
        self.sim.io.valid = 0

        self.sim.clock.tick()

        self.sim.io.valid = 1

        self.sim.clock.tick()

        self.sim.io.valid = 0

        self.sim.clock.tick()

        # HW write
        # sim.io.basicreg_basicfield_wdata = literal_eval('0xB16B00B5')
        # sim.io.basicreg_basicfield_we = 0
        # sim.clock.tick()
        # sim.io.basicreg_basicfield_we = 1
        # sim.clock.tick()

        if verify:
            assert int(self.hifread(addr))==int(val)

    def hifread(self, addr='0x00'):
        """ Read register """
        addr = hifread_preproc(addr)

        # SW read
        self.sim.io.addr = addr
        self.sim.io.read = 1

        self.sim.io.valid = 0
        self.sim.clock.tick()

        self.sim.io.valid = 1
        self.sim.clock.tick()

        outval = self.sim.io.rdata

        self.sim.io.valid = 0
        self.sim.clock.tick()

        # print(hex(outval))
        return outval


def verilator_version():
    """ Return verilator version """
    result = subprocess.run(['verilator', '--version'],
                            stdout=subprocess.PIPE, check=False)
    ver = result.stdout.split()[1]
    return ver.decode("utf-8")


def verilator_version_ok():
    """ True if the installed verilator version works as transport for cheap_pie """
    # check Verilated::flushCall() exist
    # https://github.com/chipsalliance/chisel3/issues/1565
    ver = verilator_version()
    # print(ver)

    return (
        version.parse(ver) < version.parse("4.036") or
        version.parse(ver) > version.parse("4.102")
    )


def test_cp_pyverilator(args=[]):  # pylint: disable=W0102
    """ Test pyverilator transport """
    if True:  # verilator_version_ok(): pylint: disable=W0125
        prms = cli(args)
        hif = CpPyverilatorTransport(prms.fname)
        val = literal_eval('0x5A5A5A5A')
        hif.hifwrite(val=val)
        print(hex(hif.hifread()))
        assert hif.hifread() == val
    else:
        # this was fixed by pyverilator-mm
        print('Warning: pyverilator not working anymore \
              with verilator versions between 4.036 and 4.102.')
        print('https://github.com/chipsalliance/chisel3/issues/1565')


if __name__ == '__main__':
    test_cp_pyverilator(sys.argv[1:])



#!/usr/bin/env python3
""" Cheap Pie native module parser for IP-XACT """
# this file is part of cheap_pie, a python tool for chip validation
# author: Marco Merlin
# email: marcomerli@gmail.com

from ast import literal_eval
import untangle  # for parsing xml

# import sys     # pylint: disable=C0411
# import os.path # pylint: disable=C0411
# sys.path.append( os.path.join(os.path.dirname(__file__), '..') )

from cheap_pie.cheap_pie_core.cp_builder import CpHalBuilder  # pylint: disable=C0413,E0401
from cheap_pie.cheap_pie_core.cp_cli import cp_devices_fname


def ipxact_remove_prefix(ipx):
    """ remove ipxact or spirit prefix"""
    replist = ['ipxact_', 'spirit_']

    if hasattr(ipx, 'children'):
        for elm in ipx.children:
            # remove all suffixes
            for repl in replist:
                elm._name = elm._name.replace( # pylint: disable=W0212
                    repl, '')
            # remove recursively
            ipxact_remove_prefix(elm)

    return ipx


def ipxact_parse(fname, hif=None, base_address_offset="0x00000000"):
    """ Cheap Pie native module parser for IP-XACT """
    ## read input file ########################################################
    csv = untangle.parse(fname)

    ## remove ipxact/spirit prefixes ##########################################
    csv = ipxact_remove_prefix(csv)

    ## loop over lines ########################################################
    cpb = CpHalBuilder(hif)

    periph = csv.component.memoryMaps.memoryMap.addressBlock
    base_addr_str = periph.baseAddress.cdata.replace("'h", '0x')
    base_address = literal_eval(base_addr_str)

    if hasattr(periph, 'register'):  # pylint: disable=R1702
        for reg in periph.register:
            if hasattr(reg.name, 'cdata'):
                # new register
                addr_str = reg.addressOffset.cdata.replace("'h", '0x')
                regaddr = literal_eval(
                    addr_str) + base_address + literal_eval(base_address_offset)

                cpb.reg_open(regname=f'{periph.name.cdata}_{reg.name.cdata}',
                             regaddr=regaddr)

                # for field_idx in range(nfields):
                if hasattr(reg, 'field'):
                    for field in reg.field:
                        if not field is None:
                            # Create new field
                            cpb.newfield(
                                regfield=field.name.cdata,
                                width=field.bitWidth.cdata,
                                offset=field.bitOffset.cdata,
                                comments=""  # field.description.cdata
                            )

    # convert output dictionary into structure
    return cpb.out()


def test_ipxact_parse():
    """ Test function for cheap pie native IP-XACT parser """

    ipxact_parse(cp_devices_fname("my_subblock.xml"))
    ipxact_parse(cp_devices_fname("leon2_creg.xml"))
    ipxact_parse(cp_devices_fname("generic_example.xml"))


if __name__ == '__main__':
    import sys
    if len(sys.argv) > 1:
        LFNAME = sys.argv[1]
    else:
        LFNAME = cp_devices_fname("my_subblock.xml")
    print(LFNAME)
    print(ipxact_parse(LFNAME))
